/*-
 * Copyright (c) 2004 Norikatsu Shigemura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: linux_signal.c,v 1.4 2005/05/06 16:21:33 nork Exp $
 */

/*
 *
 */
#define	pthread_sigmask	bsd_pthread_sigmask
#include "pluginwrapper.h"
#include "linux.h"
#include <errno.h>
#include <stddef.h>
#include <pthread.h>
#include <sys/param.h>
#include <sys/signal.h>
#include <sys/signalvar.h>
#undef	pthread_sigmask


static int linux_to_bsd_signal[LINUX_SIGTBLSZ] = {
	SIGHUP,  SIGINT,    SIGQUIT, SIGILL,
	SIGTRAP, SIGABRT,   SIGBUS,  SIGFPE,
	SIGKILL, SIGUSR1,   SIGSEGV, SIGUSR2,
	SIGPIPE, SIGALRM,   SIGTERM, SIGBUS,
	SIGCHLD, SIGCONT,   SIGSTOP, SIGTSTP,
	SIGTTIN, SIGTTOU,   SIGURG,  SIGXCPU,
	SIGXFSZ, SIGVTALRM, SIGPROF, SIGWINCH,
	SIGIO,   SIGURG,    SIGSYS
};


static int bsd_to_linux_signal[LINUX_SIGTBLSZ] = {
	LINUX_SIGHUP,  LINUX_SIGINT,    LINUX_SIGQUIT, LINUX_SIGILL,
	LINUX_SIGTRAP, LINUX_SIGABRT,   0,             LINUX_SIGFPE,
	LINUX_SIGKILL, LINUX_SIGBUS,    LINUX_SIGSEGV, LINUX_SIGSYS,
	LINUX_SIGPIPE, LINUX_SIGALRM,   LINUX_SIGTERM, LINUX_SIGURG,
	LINUX_SIGSTOP, LINUX_SIGTSTP,   LINUX_SIGCONT, LINUX_SIGCHLD,
	LINUX_SIGTTIN, LINUX_SIGTTOU,   LINUX_SIGIO,   LINUX_SIGXCPU,
	LINUX_SIGXFSZ, LINUX_SIGVTALRM, LINUX_SIGPROF, LINUX_SIGWINCH,
	0,             LINUX_SIGUSR1,   LINUX_SIGUSR2
};


static void
linux_to_bsd_sigset(sigset_t *sa_mask, uint32_t lsa_mask)
{
	int b_sig, l_sig;

	SIGEMPTYSET(*sa_mask);

	sa_mask->__bits[0] = lsa_mask & ~((1U << LINUX_SIGTBLSZ) - 1);
	sa_mask->__bits[1] = 0;

	for(  l_sig = 1;  l_sig <= LINUX_SIGTBLSZ;  l_sig++  )  {
		if(  LINUX_SIGISMEMBER(lsa_mask, l_sig)  )  {
			if(  (b_sig = linux_to_bsd_signal[_SIG_IDX(l_sig)])  )
				SIGADDSET(*sa_mask, b_sig);
		}
	}
}


static uint32_t
bsd_to_linux_sigset(sigset_t *sa_mask)
{
	int b_sig, l_sig;
	uint32_t lsa_mask;

	LINUX_SIGEMPTYSET(lsa_mask);

	lsa_mask = sa_mask->__bits[0] & ~((1U << LINUX_SIGTBLSZ) - 1);

	for(  b_sig = 1;  b_sig <= LINUX_SIGTBLSZ;  b_sig++  )  {
		if(  SIGISMEMBER(*sa_mask, b_sig)  )  {
			if(  (l_sig = bsd_to_linux_signal[_SIG_IDX(b_sig)])  )
				LINUX_SIGADDSET(lsa_mask, l_sig);
		}
	}

	return lsa_mask;
}


/*
 * raise(3)
 */
extern int _raise(int);

int
raise(int l_sig)
{
	int sig = linux_to_bsd_signal[_SIG_IDX(l_sig)];
	int ret;

	CLEAR_ERRNO;

	ret = _raise(sig);

	dprintf("raise(%d(%d on linux)) = %d / ERRNO = %d", sig, l_sig, ret, errno);

	return ret;
}


/*
 * kill(2) system call
 */
extern int _kill(pid_t, int);

int
kill(pid_t pid, int l_sig)
{
	int sig = linux_to_bsd_signal[_SIG_IDX(l_sig)];
	int ret;

	CLEAR_ERRNO;

	ret = _kill(pid, sig);

	dprintf("kill(%d, %d(%d on linux)) = %d / ERRNO = %d", pid, sig, l_sig, ret, errno);

	return ret;
}


/*
 * sigaction(2) system call
 */
extern int _sigaction(int, const struct sigaction*, struct sigaction*);

int
sigaction(int l_sig, const struct linux_sigaction *lact, struct linux_sigaction *olact)
{
	int ret, sig = linux_to_bsd_signal[_SIG_IDX(l_sig)];
	struct sigaction act, oact;

	CLEAR_ERRNO;

	if(  l_sig <= 0  ||  LINUX_NSIG < l_sig  )  {
		errno = EINVAL;
		ret = -1;
	}  else  {
		if(  lact != NULL  )  {
			act.sa_handler = lact->lsa_handler;
			(void)linux_to_bsd_sigset(&act.sa_mask, lact->lsa_mask);
			act.sa_flags   = 0;
			if(  lact->lsa_flags & LINUX_SA_NOCLDSTOP  )
				act.sa_flags |= SA_NOCLDSTOP;
			if(  lact->lsa_flags & LINUX_SA_NOCLDWAIT  )
				act.sa_flags |= SA_NOCLDWAIT;
			if(  lact->lsa_flags & LINUX_SA_SIGINFO  )
				act.sa_flags |= SA_SIGINFO;
			if(  lact->lsa_flags & LINUX_SA_ONSTACK  )
				act.sa_flags |= SA_ONSTACK;
			if(  lact->lsa_flags & LINUX_SA_RESTART  )
				act.sa_flags |= SA_RESTART;
			if(  lact->lsa_flags & LINUX_SA_ONESHOT  )
				act.sa_flags |= SA_RESETHAND;
			if(  lact->lsa_flags & LINUX_SA_NOMASK  )
				act.sa_flags |= SA_NODEFER;
		}

		ret = _sigaction(sig, (lact ? &act : NULL), &oact);

		if(  olact != NULL  )  {
			olact->lsa_handler  = oact.sa_handler;
			olact->lsa_restorer = NULL;
			olact->lsa_mask     = bsd_to_linux_sigset(&oact.sa_mask);
			olact->lsa_flags    = 0;
			if(  oact.sa_flags & SA_NOCLDSTOP  )
				olact->lsa_flags |= LINUX_SA_NOCLDSTOP;
			if(  oact.sa_flags & SA_NOCLDWAIT  )
				olact->lsa_flags |= LINUX_SA_NOCLDWAIT;
			if(  oact.sa_flags & SA_SIGINFO  )
				olact->lsa_flags |= LINUX_SA_SIGINFO;
			if(  oact.sa_flags & SA_ONSTACK  )
				olact->lsa_flags |= LINUX_SA_ONSTACK;
			if(  oact.sa_flags & SA_RESTART  )
				olact->lsa_flags |= LINUX_SA_RESTART;
			if(  oact.sa_flags & SA_RESETHAND  )
				olact->lsa_flags |= LINUX_SA_ONESHOT;
			if(  oact.sa_flags & SA_NODEFER  )
				olact->lsa_flags |= LINUX_SA_NOMASK;
		}
	}

	dprintf("sigaction(%d(%d on linux), 0x%p, 0x%p) = %d / ERRNO = %d", sig, l_sig, lact, olact, ret, errno);

	return ret;
}


/*
 * pthread_sigmask(3)
 */
int _pthread_sigmask(int, const sigset_t *, sigset_t *);

int
pthread_sigmask(int how, const l_sigset_t *newmask, l_sigset_t *oldmask)
{
	int ret;

	CLEAR_ERRNO;

	ret = _pthread_sigmask(how, (const sigset_t *)newmask, (sigset_t *)oldmask);

	dprintf("pthread_sigmask(%d(%d on linux), %p, %p) = %d / ERRNO = %d", how, newmask, oldmask, ret, errno);

	return ret;

#if 0
	if (newmask != NULL) {
		mask = *newmask;
		/* Don't allow __pthread_sig_restart to be unmasked.
		   Don't allow __pthread_sig_cancel to be masked. */
		switch(how) {
		case SIG_SETMASK:
			sigaddset(&mask, __pthread_sig_restart);
			sigdelset(&mask, __pthread_sig_cancel);
			break;
		case SIG_BLOCK:
			sigdelset(&mask, __pthread_sig_cancel);
			break;
		case SIG_UNBLOCK:
			sigdelset(&mask, __pthread_sig_restart);
			break;
		}
		newmask = &mask;
	}
	if (sigprocmask(how, newmask, oldmask) == -1)
		return errno;
	else
		return 0;
#endif
}


/*
 * signal(3)
 */
#if 0
extern sigset_t _sigintr;

void (*
signal( int l_sig, void (*func)(int) )
     )(int)
{
	int sig = linux_to_bsd_signal[_SIG_IDX(l_sig)];
	struct sigaction sa, osa;

	sa.sa_handler = func;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if(  sigismember(&_sigintr, sig) == 0  )
		sa.sa_flags |= SA_RESTART;
	if(  _sigaction(sig, &sa, &osa) < 0  )
		osa.sa_handler = SIG_ERR;

	dprintf("signal(%d(%d on linux), ...) / ERRNO = %d", sig, l_sig, errno);

	return osa.sa_handler;
}
#endif
